importPackage(java.lang);
importPackage(java.util);
importPackage(java.io);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLMessage);
importClass(Packages.com.alarmpoint.integrationagent.apxml.APXMLToken);
importClass(Packages.com.thoughtworks.xstream.XStream);
importClass(Packages.com.thoughtworks.xstream.converters.reflection.PureJavaReflectionProvider);
importClass(Packages.com.alarmpoint.integrationagent.util.EventDeduplicator);

var xStream = new XStream(new PureJavaReflectionProvider())

load("integrationservices/tivoli-ccmdb/baseclass.js");
load("integrationservices/tivoli-ccmdb/logger.js");
load("integrationservices/tivoli-ccmdb/wsutil.js");
load("integrationservices/tivoli-ccmdb/ccmdb.js");
load("integrationservices/tivoli-ccmdb/xmattersws.js");
load("integrationservices/tivoli-ccmdb/xMattersEventWS.js");


var exceptionString = "JavaException: ";

// Maximo Web Service Credentials
var TIVOLI_URL = "http://localhost/meaweb/services/";
var TIVOLI_USER = "maxadmin";
var TIVOLI_PASSWORD = "maxadmin";

var notePrefix = "[xMatters] - "; // The Prefix which is added to the note comments for AlarmPoint e.g. "[xMatters] "

var TOKEN_NAME = "apiaResult";
var TOKEN_NAME_EXCEPTION = "apiaResultException";

var log = new Logger();

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by APClient.  The messages may be requests to perform
 * local activity, or they may be requests to make submissions to AlarmPoint.
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * via the Integration Service's outbound queues.
 */
function apia_input(apxml)
{
  // Enrich the APXML for a workflow assignment with the change information.
  if ("WFASSIGNMENT".equals(apxml.getValue("publishchannel")))
  {
    var xmChange = new CCMDB(TIVOLI_URL+"XMCHANGE", TIVOLI_USER, TIVOLI_PASSWORD);
    var changeDetails = xmChange.queryChangeAssignment(apxml.getValue("as_ownerid"));
    apxml = xmChange.enrichAPXMLForChangeAssignment(apxml, changeDetails);

    var xmTask = new CCMDB(TIVOLI_URL+"XMTASK", TIVOLI_USER, TIVOLI_PASSWORD);
    var taskDetails = xmTask.queryWorkflowTask(apxml.getValue("as_app"), apxml.getValue("as_nodeid"), apxml.getValue("as_processname"), apxml.getValue("as_processrev"));
    apxml = xmTask.enrichAPXMLForWFTask(apxml, taskDetails);
    
    return apxml;
  }

  // Invoke event de-duplication for any other event
  // if we haven't got more than 1 event in the programmed timeout
  var recentlyReceived = EventDeduplicator.getInstance().recentOccurrenceCount(apxml, "tivoliccmdb");
  if (recentlyReceived <= 1)
  {
    if ("XMCHANGE".equals(apxml.getValue("publishchannel")))
    {
      // Send a terminate event into xMatters for any previous published events for
      // this change record
      handleDel(apxml.getValue("incident_id"));
    }
    return apxml; 
  }
  else
  {
    log.warn("***Deduplicator Suppressed Notification ***: An event with tokens " + convertTokensToString(apxml) + " has been injected into the " +
             apxml.getValue("agent_client_id") + " event domain " + recentlyReceived + " times within the configured suppression period.  It has been suppressed."
    );
  }
}

/**
 * This is the main method for handling APXML messages sent to the
 * Integration Service by AlarmPoint.  The messages may be responses to
 * previous submissions, or they may be requests to the Integration Service
 * that originate from AlarmPoint
 * (e.g., via an ExternalServiceMessage/ExternalServiceRequest2).
 * <p>
 * Any APXMLMessage object that this method returns will be sent to AlarmPoint
 * either via the Integration Service's outbound queues or directly as a
 * Web Service response, depending on the mechanism used to submit the APXML.
 */
function apia_response(apxml)
{
  log.debug("About to start handling apia_response");
  var method = apxml.getMethod();

  if (APXMLMessage.OK_METHOD.equalsIgnoreCase(method))
  {
    return handleOK(apxml);
  }
  else if (APXMLMessage.ERROR_METHOD.equalsIgnoreCase(method))
  {
    return handleERROR(apxml);
  }
  else if (APXMLMessage.SEND_METHOD.equalsIgnoreCase(method))
  {
    return handleSend(apxml);
  }
  else if (APXMLMessage.REQUEST_METHOD.equalsIgnoreCase(method))
  {
    return handleRequest(apxml);
  }
  else
  {
    throw new IllegalArgumentException("Unrecognized APXML method: " + method);
  }
}

/**
 * OK responses are returned by AlarmPoint for each APXML message that it accepts
 * from an Integration Service.
 * <p>
 * Log the acknowledgement using the ping Integration Service's logger.
 */
function handleOK(apxml)
{
  var id = apxml.getTransactionId();
  log.info("Submission of transaction " + id + " succeeded.");
}

/**
 * ERROR responses are returned by AlarmPoint for each APXML message that it
 * is unable to accept from an Integration Service.
 * <p>
 * Log the error using the ping Integration Service's logger.
 */
function handleERROR(apxml)
{
  var id = apxml.getTransactionId();
  var code = apxml.getValue(APXMLMessage.ERROR_CODE);
  log.error("Submission of transaction " + id + "failed with error code " + code);
}

/**
 * The main routine for handling Requests from AlarmPoint
 * <p>
 * Redirects to the handleSend as messages may come from ESM or ESR to the IA.
 */
function handleRequest(apxml)
{
  return handleSend(apxml);
}

/**
 * The main routine for handling Requests AlarmPoint
 * <p>
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleSend(apxml)
{
  log.debug( "handleSend received apxml: " + apxml );

  // handle annotations coming from callout APS script
  // they are formatted differently
  // the presence of the message token means this message
  // is a callout annotation.
  var message = apxml.getValue("message");
  var incidentId = apxml.getToken("incident_id").getValue();

  if ( message != null && incidentId != null)
  {
    return handleCalloutAnnotate( apxml );
  }
  else
  {
    return handleApsResponses( apxml );
  }
}

/**
 * Throw an exception that will be logged by the Integration Agent.
 */
function handleApsResponses(apxml)
{
  // ////////////////////////////////////////////////////////////////////////
  //
  // The main routine for handling requests starts here.
  //
  // ////////////////////////////////////////////////////////////////////////
  // Used to update Tivoli CCMDB Workflow using web services
  var response = initializeResponse(apxml);
  var result = null;
  var xmAssignment = new CCMDB(TIVOLI_URL+"XMASSIGNMENT", TIVOLI_USER, TIVOLI_PASSWORD);
  var xmChange = new CCMDB(TIVOLI_URL+"XMCHANGE", TIVOLI_USER, TIVOLI_PASSWORD);

  var requestText = "";
  var messageText = notePrefix + "Updated Task";
  if(apxml.getToken("request_text") != null)
  {
    requestText = apxml.getToken("request_text").getValue();
  }
  if(apxml.getToken("message_text") != null && !"".equals(apxml.getToken("message_text")))
  {
    messageText = notePrefix + apxml.getToken("message_text").getValue();
  }

  // deserialize the additional tokens
  var additionalTokens = getAdditionalTokens(apxml);
  log.debug( "additionalTokens keys: " + additionalTokens);
  var wonum = processAdditionalTokens(additionalTokens, "wonum", true);
  var siteId = processAdditionalTokens(additionalTokens, "siteid", true);
  var ownerId = processAdditionalTokens(additionalTokens, "ownerid", true);
  var responder = processAdditionalTokens(additionalTokens, "responder", true);

  var app = processAdditionalTokens(additionalTokens, "app", true);
  var assignId = processAdditionalTokens(additionalTokens, "assignid", true);
  var nodeId = processAdditionalTokens(additionalTokens, "nodeid", true);
  var processName = processAdditionalTokens(additionalTokens, "processname", true);
  var processRev = processAdditionalTokens(additionalTokens, "processrev", true);
  var wfAssignmentId = processAdditionalTokens(additionalTokens, "wfassignmentid", true);
  var wfId = processAdditionalTokens(additionalTokens, "wfid", true);
  
  try
  {
    // For Annotate Request
    if ( requestText.equalsIgnoreCase( "Annotate" ))
    {
      if (!"".equalsIgnoreCase(siteId))
      {
        xmChange.sendAnnotateRequest(siteId, ownerId, wonum, 0, messageText, "CLIENTNOTE");
      }
    }
    // For Accept Request
    else if ( requestText.equalsIgnoreCase( "Accept" ) || requestText.equalsIgnoreCase( "Reject" ) )
    {
      var actionId = processAdditionalTokens(additionalTokens, "actionid", true);
      var memo = processAdditionalTokens(additionalTokens, "memo", true);
      
      xmAssignment.sendCompleteAssignmentRequest(app, assignId, nodeId, processName, processRev, wfAssignmentId, wfId, actionId, memo);
    }
    else if ( requestText.equalsIgnoreCase( "Reassign" ) )
    {
      var assignee = processAdditionalTokens(additionalTokens, "assignee", true);
      var memo = processAdditionalTokens(additionalTokens, "memo", true);
      var taskDesc = processAdditionalTokens(additionalTokens, "taskdesc", true);
      
      xmAssignment.sendReassignRequest(app, assignId, nodeId, processName, processRev, wfAssignmentId, wfId, assignee, memo, 1, taskDesc);
    }
    else
    {
      throw new Exception("Request Script Failed: RequestText: [" + requestText + "] not implemented");
    }
    log.debug("Result of "+requestText+" returns "+result);
    response.setToken(TOKEN_NAME, "OK", APXMLToken.Type.STRING);
  }
  catch (e)
  {
    response.setToken(TOKEN_NAME, "OK", APXMLToken.Type.STRING);
    var message = "Maximo could not process the request [" + requestText + "] and returned the following message: ";
    if (e.toString().indexOf("interaction") != -1)
    {
      message += "The workflow could not be completed for Change " + wonum + " due to a required interaction. Please logon to Maximo to complete the workflow.";
    }
    else
    {
      log.error("Caught exception processing [" + requestText + "] Exception:" + e);
      message += e.toString().replace(exceptionString, "");
    }
    message += " To resolve this issue, please review the message or contact your Tivoli Administrator for assistance.";
    sendQuickMessage(assignId, responder, message);
  }
  log.debug("Finished call to management system for requestText: [" + requestText + "]");
  log.debug("Response: " + response);

  return response;
}

/**
 * handles annotations sent from the Callout APS script
 */
function handleCalloutAnnotate( apxml )
{
  var message = apxml.getValue("message");

  var additionalTokens = getAdditionalTokens(apxml);
  log.debug( "additionalTokens keys: " + additionalTokens);
  var wonum = processAdditionalTokens(additionalTokens, "wonum", true); 
  var siteId = processAdditionalTokens(additionalTokens, "siteid", true); 
  var ownerId = processAdditionalTokens(additionalTokens, "ownerid", true); 

  log.debug( "About to call to management system for callout annotate for [" + wonum + "]" );

  try
  {
    if (!"".equalsIgnoreCase(siteId))
    {
      var xmChange = new CCMDB(TIVOLI_URL+"XMCHANGE", TIVOLI_USER, TIVOLI_PASSWORD);
      xmChange.sendAnnotateRequest(siteId, ownerId, wonum, 0, message, "CLIENTNOTE");
    }
  }
  catch (e)
  {
    log.error("Caught exception processing annotation for [" + wonum + "] Exception:" + e);
  }

  log.debug("Finished call to management system for callout annotate");
}

/**
 * Initializes the response to be returned to the device
 * @param apxml APXML containing the request information
 * @return initialized instance of the response APXML
 */
function initializeResponse(apxml)
{
  var response = ServiceAPI.createAPXML();
  response.setMethod(APXMLMessage.RESPONSE_METHOD);
  response.setSubclass(APXMLMessage.RESPONSE_SUBCLASS);

  if (apxml.hasToken(APXMLMessage.REQUEST_ID))
  {
    response.setToken(APXMLMessage.REQUEST_ID, apxml.getValue(APXMLMessage.REQUEST_ID), APXMLToken.Type.STRING);
  }
  response.setToken(APXMLMessage.DESTINATION, apxml.getValue(APXMLMessage.ORIGINATOR), APXMLToken.Type.STRING);
  if (apxml.hasToken(APXMLMessage.APIA_PRIORITY))
  {
    response.setToken(APXMLMessage.APIA_PRIORITY, apxml.getValue(APXMLMessage.APIA_PRIORITY));
  }
  else
  {
    response.setToken(APXMLMessage.APIA_PRIORITY, APXMLMessage.APIA_PRIORITY_HIGH);
  }
  return response;
}

/**
 * Retrieves the additional tokens from the APXML and throws an exception if a mandatory token is missing
 * @param additionalTokens collection of tokens to process
 * @param tokenName name of the token to retrieve
 * @param mandatory true if token is required, false otherwise
 * @return the token value
 */
function processAdditionalTokens(additionalTokens, tokenName, mandatory)
{
  var token = additionalTokens.get(tokenName);
  if (mandatory && token == null)
  {
    throw new Exception("additionalTokens in apxml is missing " + tokenName + " which is mandatory");
  }
  return token;
}

/**
 * Retrieves the additional tokens from the APXML
 * @param apxml XML message to process
 * @return additional tokens
 * @throws if the additional tokens are missing from the XML.
 */
function getAdditionalTokens(apxml)
{
  if (apxml.getValue("additionalTokens") != null && !"".equalsIgnoreCase(apxml.getValue("additionalTokens"))) 
  {
    return xStream.fromXML(apxml.getValue("additionalTokens")); // LinkedHashMap
  }
  throw new Exception("additionalTokens is missing from apxml"); 
}

/**
 * Sends a quick message to xMatters
 * @param incidentId id of the incident the quick message refers to
 * @param recipient id of the recipient
 * @param message string message to send
 */
function sendQuickMessage(incidentId, recipient, message)
{
  var apxml = ServiceAPI.createAPXML();
  apxml.setMethod("ADD");
  apxml.setSubclass("action");
  apxml.setToken(APXMLMessage.APIA_PROCESS_GROUP, incidentId);
  apxml.setToken("incident_id", incidentId);
  apxml.setToken("subject", "IBM Tivoli CCMDB Exception Notification");
  apxml.setToken("recipients", recipient);
  apxml.setToken("message", message);
  apxml.setToken("fyi", "true");

  log.debug("sendQuickMessage apxml: " + apxml);
  
  var response = initializeResponse(apxml);
  try
  {
    ServiceAPI.sendAPXML(apxml);      
    response.setToken(TOKEN_NAME, "OK", APXMLToken.Type.STRING);
    log.debug("Finished sending apxml to AlarmPoint");
  }
  catch (e)
  {
    log.error("Caught exception processing [" + requestText + "] Exception:" + e);
    response.setToken(TOKEN_NAME_EXCEPTION, e.toString().replace(exceptionString, ""), APXMLToken.Type.STRING);
  }

  log.debug(response);
}

/**
 * The main routine for handling del injections from the management system to xMatters
 * @param incidentId id of the incident event that will be terminated
 */  
function handleDel(incidentId)
{
  var xmEventWS = new xMattersEventWS();
  // Check if the notification exists in xMatters
  if (xmEventWS.queryIncident(incidentId))
  {
    var apxml = ServiceAPI.createAPXML();
    apxml.setMethod("DEL");
    apxml.setSubclass("action");
    apxml.setToken(APXMLMessage.APIA_PROCESS_GROUP, incidentId);
    apxml.setToken("incident_id", incidentId);

    log.debug("handleDel apxml: " + apxml);
    
    var response = initializeResponse(apxml);
    try
    {
      ServiceAPI.sendAPXML(apxml);      
      response.setToken(TOKEN_NAME, "OK", APXMLToken.Type.STRING);
      log.debug("Finished sending apxml to AlarmPoint");
    }
    catch (e)
    {
      log.warn("Unable to terminate event " + incidentId);
    }

    log.debug(response);
  }
}

/**
 * Convert all tokens for this message into a parenthesis-grouped set of name-value-pairs.
 */
function convertTokensToString(apxml)
{
  var string = "";
  var tokenIterator = apxml.getTokens();
  while (tokenIterator.hasNext())
  {
    var token = tokenIterator.next();
    string += "(" + token.getKey() + ", " + token.getValue() + ")";
  }
  return string;
}